from unittest import TestCase
from yo_fluq import *

class ExtendedMethodsTests(TestCase):
    def test_foreach(self):
        buf = []
        Query.args(1, 2, 3).foreach(buf.append)
        self.assertListEqual([1,2,3],buf)

    def test_foreach_and_continue(self):
        buf = []
        Query.args(1, 2, 3).foreach_and_continue(lambda z: buf.append(z + 10)).foreach(buf.append)
        self.assertListEqual([11,1,12,2,13,3], buf)

    def test_append(self):
        self.assertListEqual(
            [1,2,5,6],
            Query.args(1, 2).append(5, 6).to_list()
        )

    def test_prepend(self):
        self.assertListEqual(
            [5, 6, 1, 2],
            Query.args(1, 2).prepend(5, 6).to_list()
        )


from unittest import TestCase
from yo_fluq import *

class CoreTestCase(TestCase):
    def test_simple_aggregation(self):
        self.assertEqual(
            2.0,
            Query.push().mean()([1,2,3])
        )

    def test_select(self):
        self.assertListEqual(
            [10,20,30],
            Query.push().select(lambda z: z*10).to_list()([1,2,3])
        )

    def test_select_many(self):
        self.assertListEqual(
            ['a','b','c','d','e','f'],
            Query.push().select_many(lambda z: z).to_list()(['abc','','de','f'])
        )

    def test_where(self):
        self.assertListEqual(
            [1,2],
            Query.push().where(lambda z: z<3).to_list()([1,2,3,4,5])
        )

    def test_split_group(self):
        self.assertDictEqual(
            {
                0: [2,4],
                1: [1,3]
            },
            Query.push().split_by_group(lambda z: z%2).to_list()([1,2,3,4])
        )

    def test_split_group_always_dict(self):
        self.assertDictEqual(
            {
                0: [0]
            },
            Query.push().split_by_group(lambda z: z).to_list()([0])
        )

    def test_split_group_with_total(self):
        self.assertDictEqual(
            {
                0: [2,4],
                1: [1,3],
                'total': [1,2,3,4]
            },
            Query.push().split_by_group(lambda z: z%2,'total').to_list()([1,2,3,4])
        )

    def test_split_dict(self):
        self.assertDictEqual(
            dict(
                one = 2,
                ten = 20,
                hundred = 200
            ),
            Query.push().select(lambda z: dict(one=z,ten=z*10, hundred=z*100)).split_dictionary().sum()([1,1])
        )

    def test_split_dict_raise(self):
        self.assertRaises(
            ValueError,
            lambda: Query.push().split_dictionary().mean()([1])
        )



    def test_split_pipelines(self):
        self.assertDictEqual(
            dict(odds=[1,3], evens=[2,4]),
            Query.push().split_pipelines(
                odds = Query.push().where(lambda z: z%2==1).to_list(),
                evens = Query.push().where(lambda z: z%2==0).to_list(),
            )([1,2,3,4])
        )

    def test_subscription_raises_by_leaves(self):
        self.assertRaises(
            ValueError,
            lambda: Query.push().mean().mean()
        )

    def test_empty_pq_raises(self):
        self.assertRaises(
            ValueError,
            lambda: Query.push()([1,2])
        )

    def test_empty_observable(self):
        self.assertIsNone(
            Query.push().select(lambda z: z)([1,2])
        )



from . import aggregation as agg
from .transformation import SplitPipelines
from typing import *

class _NoDefault:
    pass

class AggregationCodeFactory:
    def __init__(self, method_concatenator: Callable):
        self.method_concatenator = method_concatenator

    def first(self):
        return self.method_concatenator(agg.First())

    def first_or_default(self, default=None):
        return self.method_concatenator(agg.First().default_if_empty(default))

    def last(self):
        return self.method_concatenator(agg.Last())

    def last_or_default(self, default=None):
        return self.method_concatenator(agg.Last().default_if_empty(default))

    def single(self):
        return self.method_concatenator(agg.Single())

    def single_or_default(self, default=None):
        return self.method_concatenator(agg.Single().default_if_empty(default))

    def min(self, selector : Callable = lambda z: z, default = _NoDefault()):
        a = agg.Min(selector)
        if not isinstance(default,_NoDefault):
            a = a.default_if_empty(default)
        return self.method_concatenator(a)

    def max(self, selector: Callable = lambda z: z, default = _NoDefault()):
        a = agg.Max(selector)
        if not isinstance(default,_NoDefault):
            a = a.default_if_empty(default)
        return self.method_concatenator(a)

    def argmin(self, selector: Callable = lambda z: z, default = _NoDefault()):
        a = agg.ArgMin(selector)
        if not isinstance(default,_NoDefault):
            a = a.default_if_empty(default)
        return self.method_concatenator(a)

    def argmax(self, selector: Callable = lambda z: z, default = _NoDefault()):
        a = agg.ArgMax(selector)
        if not isinstance(default, _NoDefault):
            a = a.default_if_empty(default)
        return self.method_concatenator(a)

    def sum(self):
        return self.method_concatenator(agg.Sum())

    def mean(self):
        return self.method_concatenator(agg.Mean())

    def count(self):
        return self.method_concatenator(agg.Count())

    def any(self, selector: Optional[Callable] = None):
        return self.method_concatenator(agg.Any(selector))

    def all(self, selector: Optional[Callable] = None):
        return self.method_concatenator(agg.All(selector))

    def to_list(self):
        return self.method_concatenator(agg.ToList())

    def to_dictionary(self, key_selector: Optional[Callable] = None, value_selector: Optional[Callable] = None):
        return self.method_concatenator(agg.ToDictionary(key_selector, value_selector))

    def to_set(self):
        return self.method_concatenator(agg.ToSet())

    def to_tuple(self):
        return self.method_concatenator(agg.ToTuple())

    def aggregate_with(self, *args: agg.PushQueryElement):
        if len(args)==1:
            return self.method_concatenator(args[0])
        else:
            return self.method_concatenator(SplitPipelines(*args))



from unittest import TestCase
from yo_fluq import *

class TestListAggreagation(TestCase):
    def test_aggregate_with(self):
        self.assertEqual(
            3,
            Query.args(1,2,3).aggregate_with(agg.Count())
        )

    def test_aggregate_list(self):
        self.assertDictEqual(
            dict(count=3,mean=2.0),
            Query.args(1,2,3).aggregate_with(agg.Count(),agg.Mean())
        )

    def test_throw_if_empty(self):
        self.assertRaises(ValueError,lambda:Query.args().aggregate_with(agg.Count().throw_if_empty()))


from unittest import TestCase
from yo_fluq import *


class GeneralSelectorsTests(TestCase):

    def test_select(self):
        self.assertListEqual(
            ['1','2'],
            Query.args(1, 2).select(str).to_list()
        )

    def test_select_length(self):
        self.assertEqual(3,Query.en([1,2,3]).select(str).length)

    def test_select_many(self):
        self.assertListEqual(
            [0, 0, 1, 0, 1, 2],
            Query.args(1, 2, 3).select_many(lambda z: range(z)).to_list())

    def test_distinct(self):
        self.assertListEqual(
            [1, 2, 3],
            Query.args(1, 2, 2, 3, 3).distinct().to_list()
        )

    def test_distinct_1(self):
        self.assertListEqual(
            ['aa','b'],
            Query.args('aa', 'ab', 'b').distinct(lambda z: z[0]).to_list()
        )

    def test_where(self):
        self.assertListEqual(
            [2,4],
            Query.args(1, 2, 3, 4).where(lambda z: z % 2 == 0).to_list()
        )

    def test_group_by(self):
        result = Query.args(2, 3, 1, 4).group_by(lambda z: z % 2).to_list()
        self.assertEqual(0, result[0].key)
        self.assertListEqual([2, 4], result[0].value)
        self.assertEqual(1, result[1].key)
        self.assertListEqual([3, 1], result[1].value)

    def test_group_by_1(self):
        result = Query.args(2,3,1,4).group_by(lambda z: z%2).select(lambda z: (z.key,Query.en(z).sum())).to_list()
        self.assertListEqual(
            [(0,6),(1,4)],
            result)

    def test_group_by_2(self):
        result = Query.args(2,3,1,4).group_by(lambda z: z%2).to_dictionary()
        self.assertDictEqual({0:[2,4],1:[3,1]},result)

    def test_aggrerate(self):
        self.assertEqual('123', Query.args('1', '2', '3').aggregate(lambda acc, el: acc + el))

    def test_concat(self):
        self.assertListEqual(
            [1,2,3,3,4,5],
            Query.args(1,2,3).concat([3,4,5]).to_list()
        )

    def test_intersect(self):
        self.assertListEqual(
            [2,3],
            Query.args(1,2,3).intersect([2,3,4]).to_list()
        )

    def test_skip_while(self):
        self.assertListEqual(
            [3,4,5],
            Query.args(1,2,3,4,5).skip_while(lambda z: z<3).to_list()
        )

    def test_take_while(self):
        self.assertListEqual(
            [1,2],
            Query.args(1,2,3,4,5).take_while(lambda z: z<3).to_list()
        )

    def test_skip(self):
        self.assertListEqual(
            [4,5],
            Query.args(1, 2, 3, 4, 5).skip(3).to_list()
        )

    def test_take(self):
        self.assertListEqual(
            [1,2],
            Query.args(1, 2, 3, 4, 5).take(2).to_list()
        )

    def test_iter(self):
        buf = []
        for e in Query.args(1,2,3):
            buf.append(e)
        self.assertListEqual([1,2,3],buf)

    def test_feed(self):
        self.assertListEqual(
            [1,2,3],
            Query.args(1,2,3).feed(list)
        )


from unittest import TestCase
from yo_fluq import KeyValuePair

class MiscTestCase(TestCase):
    def test_kv_pair_repr(self):
        kvp = KeyValuePair((2,3),[dict(a=2)])
        self.assertEqual("{'key': (2, 3), 'value': [{'a': 2}]}", kvp.__repr__())
        self.assertEqual("{'key': (2, 3), 'value': [{'a': 2}]}", kvp.__str__())



from yo_fluq import *
from unittest import TestCase

def lazy_en_for_first():
    yield 0
    raise ValueError('First is not Lazy') #pragma: no cover
    yield 1                               #pragma: no cover

class LittleMethodsTests(TestCase):


    def test_first_1(self):
        self.assertEqual(1, Query.args(1, 2, 3).first())


    def test_first_2(self):
        self.assertRaises(ValueError,lambda: Query.args().first())

    def test_first_or_default_1(self):
        self.assertEqual(1, Query.args(1, 2, 3).first_or_default())


    def test_first_or_default_2(self):
        self.assertEqual(-1, Query.args().first_or_default(-1))

    def test_first_lazy(self):
        self.assertEqual(0,Query.en(lazy_en_for_first()).first())

    def test_last_1(self):
        self.assertEqual(3, Query.args(1, 2, 3).last())


    def test_last_2(self):
        self.assertRaises(ValueError, lambda: Query.args().last())


    def test_last_or_default_1(self):
        self.assertEqual(3, Query.args(1, 2, 3).last_or_default())


    def test_last_or_default_2(self):
        self.assertEqual(-1, Query.args().last_or_default(-1))


    def test_single_1(self):
        self.assertEqual(1, Query.args(1).single())


    def test_single_2(self):
        self.assertRaises(ValueError,lambda: Query.args().last())


    def test_single_3(self):
        self.assertRaises(ValueError,lambda: Query.args(1, 2).single())


    def test_single_or_default_1(self):
        self.assertEqual(1, Query.args(1).single_or_default())


    def test_single_or_default_2(self):
        self.assertEqual(-1, Query.args().single_or_default(-1))


    def test_single_or_default_3(self):
        self.assertRaises(ValueError,lambda:Query.args(1, 2).single_or_default())




